#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

import re
import sys
import subprocess
import time
import os
import signal
import argparse
import uuid

import frama_c_results
import results_display
import benchmark_database

class OperationException(Exception):
    pass

def build_make_environment(framac):
    if framac is None:
        env = { **os.environ }
        args = []
    else:
        env = { **os.environ,  'PATH' : f"{framac}/bin:{os.environ['PATH']}" }
        args = [
            f"FRAMAC_BIN={framac}/bin",
            f"FRAMAC={framac}/bin/frama-c"
        ]
    return env, args

def list_targets(dir):
    if not os.path.isdir(dir):
        raise OperationException(f"target is not a directory: {dir}")

    env, args = build_make_environment(framac)
    res = subprocess.run(
        ["make", "--directory", dir, "--quiet", "display-targets"] + args,
        env=env,
        stdout=subprocess.PIPE,
        encoding='ascii')
    targets = res.stdout.split()
    res = []
    for target in targets:
        if target.endswith(".eva") or target.endswith(".parse"):
            res += [f"{dir}/{target}"]
        else:
            res += list_targets(target)
    return res

def clone_frama_c(clonedir, hash):
    print("Cloning Frama-C", hash, "...")
    res = subprocess.run(
        ["./scripts/clone.sh", "--clone-dir", clonedir, hash],
        stdout=subprocess.PIPE,
        encoding='ascii')
    if res.returncode != 0:
        raise OperationException("Cannot clone repository. Try to manually"
            "remove the broken clone in " + clonedir)
    return res.stdout.strip() + '/build'

def run_make(framac, benchmark_tag=None):
    args = ['make', '--keep-going', 'all']
    env, var_args = build_make_environment(framac)
    if benchmark_tag is None:
        args += ['-j', str(os.cpu_count ())]
    else:
        args += ['BENCHMARK=' + benchmark_tag]
    return subprocess.Popen(args + var_args, env=env,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        preexec_fn=os.setsid)

def terminate_process(process):
    if process is None:
        return b""
    else:
        try:
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
            pass
        except ProcessLookupError:
            pass
        output,errors = process.communicate()
        return errors

def smart_rename(target):
    target = re.sub('^\./', '', target)
    target = re.sub('main\.eva$', '', target)
    target = re.sub('\.eva$', '', target)
    target = re.sub('\.frama-c/', '', target)
    target = re.sub('qds/frama-c', 'qds', target)
    return target

def is_running(target):
    return os.path.isfile(target + '/running')

def poll_results(targets, benchmark_tag):
    results = []
    for target in targets:
        filename = target + '/stats.txt'
        result = frama_c_results.read(filename)
        result["target"] = target
        result["target_name"] = smart_rename(target)
        result["is_running"] = is_running(target)
        result["up_to_date"] = benchmark_tag is None or benchmark_tag == result['benchmark_tag']
        results.append(result);
    return results


def run_analyses(display, database, framac, benchmark_tag):
    results = []
    targets = list_targets(".")
    process = run_make(framac, benchmark_tag)
    errors = b""
    next_poll = time.time()

    def update():
        nonlocal  display, database, targets, benchmark_tag, results
        results = poll_results(targets, benchmark_tag)
        if not database is None:
            database.update(results)
        display.needs_update = True

    try:
        while process.poll() is None:
            if time.time() >= next_poll:
                update()
                next_poll = time.time() + 2.0
            display.process_inputs()
            if display.needs_update:
                display.print_table(results)
            time.sleep(0.05)
        update()
    except (KeyboardInterrupt, results_display.UserExitRequest):
        print("Analyzes interrupted by user.")
    except Exception as e:
        # terminate_process below is somehow blocking the exception printing
        errors += bytearray(str(e), 'ascii')
        raise e
    finally:
        errors += terminate_process(process)
    return results,errors


parser = argparse.ArgumentParser(
    description="Run analyses and summarize the results. Must be run in a "
                "directory with a Makefile having two rules: 'all', a target "
                "that runs the analysis, and 'display-targets', the target that "
                "lists the built results.")
parser.add_argument('rev', nargs='?', metavar="REVISION",
    help="a Frama-C revision to use for analyses (default: use the "
        "default configuration for Frama-C)")
parser.add_argument('-b', '--benchmark',
    action="store_true",
    help="sets benchmark mode: do not run analyses in parallel and rerun all "
        "analyses")
parser.add_argument('-v', '--vs',
    action="store", metavar="REVISION", default="master",
    help="a revision to compare the results to")
parser.add_argument('-c', '--comment',
    action="store", metavar="COMMENT",
    help="when benchmarking, add this comment inside the database")
parser.add_argument('-p', '--repository-path',
    action="store", metavar="PATH",
    help="don't clone Frama-C, use this git repository instead")


errors = b''

try:
    args = parser.parse_args()

    if args.repository_path is None:
        if args.rev is None:
            gitdir = None
            framac = None
        else:
            clonedir = "./frama-c-clones"
            gitdir = clonedir + "/frama-c.git"
            framac = clone_frama_c(clonedir, args.rev)
    else:
        framac = os.path.abspath(args.repository_path)
        gitdir = framac

    if args.benchmark:
        benchmark_tag=str(uuid.uuid1())
        print("Running benchmarks with benchmark tag", benchmark_tag, "...")
    else:
        benchmark_tag=None
        print("Running analyses ...")

    benchmark_comment = args.comment

    if gitdir is None:
        database = None
    else:
        database = benchmark_database.Database(benchmark_tag, benchmark_comment,
            gitdir, args.rev, args.vs)

    results,errors = results_display.wrapper(run_analyses, database, framac,
        benchmark_tag, curses=True)

    print("Results:\n")
    results_display.PlainDisplay().print_table(results)

except OperationException as e:
    errors += bytearray(str(e), 'ascii')

sys.stderr.buffer.write(errors + b'\n')



#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# This script removes absolute path references in a JSON Compilation Database.
#
# See: http://clang.llvm.org/docs/JSONCompilationDatabase.html

import sys
import os
import json
import re
from pathlib import Path

MIN_PYTHON = (3, 6) # for automatic Path conversions
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

if len(sys.argv) < 2:
   # no argument, assume default name
   arg = Path("compile_commands.json")
else:
   arg = Path(sys.argv[1])

if not arg.exists():
   print(f"error: file '{arg}' not found")
   sys.exit(f"usage: {sys.argv[0]} [compile_commands.json]")

jcdb_json = json.loads(open(arg).read())
jcdb_dir = arg.parent
out_json = {}

nb_diffs = 0
for entry in jcdb_json:
   if "file" in entry and os.path.isabs(entry["file"]):
      old_entry = entry["file"]
      entry["file"] = os.path.relpath(entry["file"], jcdb_dir)
      if old_entry != entry["file"]:
         nb_diffs += 1
      else:
         print(f"warning: absolute path could not be normalized: {entry['file']}")
   elif "directory" in entry and os.path.isabs(entry["directory"]):
      old_entry = entry["directory"]
      entry["directory"] = os.path.relpath(entry["directory"], jcdb_dir)
      if old_entry != entry["directory"]:
         nb_diffs += 1
      else:
         print(f"warning: absolute path could not be normalized: {entry['directory']}")

if nb_diffs == 0:
   print(f"No changes to be applied to {arg}")
else:
   yn = input(f"{nb_diffs} replacements to be applied. Normalize {arg}? [y/N] ")
   if yn.lower() == "y":
      with open(arg, 'w', encoding='utf-8') as outfile:
         json.dump(jcdb_json, outfile, ensure_ascii=False, indent=4)
         print(f"Normalization applied to {arg}")
   else:
      print("Exiting without overwriting.")



#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# Exports find_function_in_file, to be used by other scripts

import re

# To minimize the amount of false positives, we try to match the following:
# - the line must begin with a C identifier (declarations and definitions in C
#   rarely start with spaces in the line), or with the function name itself
#   (supposing the return type is in the previous line)
# - any number of identifiers are allowed (to allow for 'struct', 'volatile',
#   'extern', etc)
# - asterisks are allowed both before and after identifiers, except for the
#   first one (to allow for 'char *', 'struct **ptr', etc)
# - identifiers are allowed after the parentheses, to allow for some macros/
#   modifiers

# Precomputes the regex for 'fname'
def prepare(fname):
    c_identifier = "[a-zA-Z_][a-zA-Z0-9_]*"
    c_id_maybe_pointer = c_identifier + "\**"
    type_prefix = c_id_maybe_pointer + "(?:\s+\**" + c_id_maybe_pointer + ")*\s+\**"
    parentheses_suffix = "\s*\([^)]*\)"
    re_fun = re.compile("^(?:" + type_prefix + "\s*)?" + fname + parentheses_suffix
                        + "\s*(?:" + c_identifier + ")?\s*(;|{)", flags=re.MULTILINE)
    return re_fun

# Returns 0 if not found, 1 if declaration, 2 if definition
def find(prepared_re, f):
   with open(f, encoding="ascii", errors='ignore') as content_file:
      content = content_file.read()
      has_decl_or_def = prepared_re.search(content)
      if has_decl_or_def is None:
          return 0
      else:
         is_decl = has_decl_or_def.group(1) == ";"
         return 1 if is_decl else 2



##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

import re

stat_file_re = re.compile("^([^=]*)=(.*)$", re.MULTILINE)

def load(filename):
    data = {}
    try:
        with open(filename, 'r') as file:
            content = file.read()
            for (key,value) in stat_file_re.findall(content):
                data[key] = value
    except OSError:
        pass
    return data

re_escape_space = re.compile(r'\\ ')

def convert(data, key, to_type, default=None):
    try:
        value = data[key].strip()
        if to_type is str:
            value = re.sub(r'\\ ', ' ', value)
            value = re.sub(r'\\,', ',', value)
            return value
        else:
            return to_type(value)
    except (ValueError, TypeError, KeyError):
        return default

def parse(data):
    result = {}
    result["timestamp"] = convert(data, "timestamp", str)
    result["sem_reach_fun"] = convert(data, "sem_reach_fun", int)
    result["syn_reach_fun"] = convert(data, "syn_reach_fun", int)
    result["total_fun"] = convert(data, "total_fun", int)
    result["sem_reach_stmt"] = convert(data, "sem_reach_stmt", int)
    result["syn_reach_stmt"] = convert(data, "syn_reach_stmt", int)
    result["alarms"] = convert(data, "alarms", int)
    result["warnings"] = convert(data, "warnings", int)
    result["user_time"] = convert(data, "user_time", float)
    result["memory"] = convert(data, "memory", int)
    result["cmd_args"] = convert(data, "cmd_args", str)
    result["benchmark_tag"] = convert(data, "benchmark_tag", str)
    if result["sem_reach_stmt"] != None and result["syn_reach_stmt"] != None:
        result["coverage"] = result["sem_reach_stmt"] / result["syn_reach_stmt"]
    else:
        result["coverage"] = None
    return result

def read(filename):
    return parse(load(filename))



#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# This script finds files containing likely declarations and definitions
# for a given function name, via heuristic syntactic matching.

import argparse
import sys
import os
import re
import glob
import function_finder

MIN_PYTHON = (3, 5) # for glob(recursive)
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

parser = argparse.ArgumentParser(description="""
Looks for likely declarations/definitions of a function
in files with extensions '.c', '.h' and '.i'.
If any directories are specified, looks inside them,
otherwise looks inside PWD and /usr/include.
Subdirectories are always considered recursively.""")

parser.add_argument('--directory', '-C', metavar='DIR', default=".", nargs=1,
                    help='print paths relative to directory DIR (default: .)')
parser.add_argument('funcname', help='function name to search')
parser.add_argument('dir', nargs='*', help='directories where to search (if empty: PWD /usr/include)')
args = vars(parser.parse_args())

reldir = args["directory"][0]
fname = args["funcname"]
dirs = args["dir"]

if re.match('[a-zA-Z_][a-zA-Z0-9_]*$', fname) == None:
    print("error: function name contains invalid characters: %s" % fname)
    print("       (only letters/digits/underscore allowed)")
    sys.exit(1)

dirs = set(dirs)
if not dirs:
    pwd = os.getcwd()
    dirs = [pwd, "/usr/include"]
else:
    dirs = set(sys.argv[2:])

files = []
for d in dirs:
    files += glob.glob(d + "/**/*.[ich]", recursive=True)

print("Looking for '%s' inside %d file(s)..." % (fname, len(files)))

possible_declarators = []
possible_definers = []
re_fun = function_finder.prepare(fname)
for f in files:
    found = function_finder.find(re_fun, f)
    if found:
        if found == 1:
            possible_declarators.append(f)
        else:
            possible_definers.append(f)

def relative_path_to(start):
    return lambda p: os.path.relpath(p, start=start)

if possible_declarators == [] and possible_definers == []:
    print("No declaration/definition found for function '%s'" % fname)
else:
    if reldir != ".":
        reldir_msg = f" (relative to '{reldir}')"
    else:
        reldir_msg = ""
    relative_path = relative_path_to(reldir)
    if possible_declarators != []:
        print(f"Possible declarations for function '{fname}' in the following file(s){reldir_msg}:")
        print("  " + "\n  ".join([os.path.relpath(path, start=reldir) for path in possible_declarators]))
    if possible_definers != []:
        print(f"Possible definitions for function '{fname}' in the following file(s){reldir_msg}:")
        print("  " + "\n  ".join([os.path.relpath(path, start=reldir) for path in possible_definers]))



##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

import subprocess

def rev_parse(gitdir, rev):
    res = subprocess.run(
        ['git', 'rev-parse', rev],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        encoding='ascii',
        cwd=gitdir)
    name = res.stdout.strip()
    return name if name else None

def name_rev(gitdir, rev):
    res = subprocess.run(
        ['git', 'name-rev', '--name-only', rev],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        encoding='ascii',
        cwd=gitdir)
    name = res.stdout.strip()
    return name if name else None

def current_rev(gitdir):
    return name_rev(gitdir, "HEAD")

def is_clean(gitdir):
    # git diff and diff-index are not working on some of our case studies to
    # decide whether the workingin dir is clean or not ; git status is more
    # reliable
    res = subprocess.run(
        ['git', 'status', '--untracked-files=no', '--porcelain'],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        encoding='ascii',
        cwd=gitdir)
    return res.returncode == 0 and not res.stdout




#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# This script serves as wrapper to 'make' (when using the analysis-scripts
# GNUmakefile template): it parses the output and suggests useful commands
# whenever it can, by calling frama-c-script itself.

import argparse
import os
import re
import subprocess
import sys
from functools import partial

MIN_PYTHON = (3, 6) # for automatic Path conversions
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

parser = argparse.ArgumentParser(description="""
Builds the specified target, parsing the output to identify and recommend
actions in case of failure.""")
parser.add_argument('--make-dir', metavar='DIR', default=".frama-c", nargs=1,
                    help='directory containing the makefile (default: .frama-c)')

(make_dir_arg, args) = parser.parse_known_args()
make_dir = vars(make_dir_arg)["make_dir"]
args = args[1:]

framac_bin = os.getenv('FRAMAC_BIN')
if not framac_bin:
   sys.exit("error: FRAMAC_BIN not in environment (set by frama-c-script)")
framac_script = f"{framac_bin}/frama-c-script"

out = subprocess.Popen(['make', "-C", make_dir] + args,
                       stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

output = out.communicate()[0].decode('utf-8')

re_missing_spec = re.compile("Neither code nor specification for function ([^,]+),")
re_recursive_call_start = re.compile("detected recursive call")
re_recursive_call_end = re.compile("Use -eva-ignore-recursive-calls to ignore")

tips = []

lines = iter(output.splitlines())
for line in lines:
    print(line)
    match = re_missing_spec.search(line)
    if match:
       fname = match.group(1)
       def action(fname):
           out = subprocess.Popen([framac_script, "find-fun", "-C", make_dir, fname],
                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
           output = out.communicate()[0].decode('utf-8')
           re_possible_definers = re.compile("Possible definitions for function")
           find_fun_lines = iter(output.splitlines())
           for find_fun_line in find_fun_lines:
              if re_possible_definers.match(find_fun_line):
                 found_files = [next(find_fun_lines)]
                 while True:
                    try:
                       found_files.append(next(find_fun_lines))
                    except StopIteration:
                       if len(found_files) > 1:
                          print("Found several files defining function '"
                                + fname + "', cannot recommend automatically.")
                          print("Check which one is appropriate and add it " +
                                "to the list of sources to be parsed:")
                          print("\n".join(found_files))
                       else:
                          print("Add the following file to the list of "
                                + "sources to be parsed:\n" + found_files[0])
                       return
           print("Could not find any files defining " + fname + ".")
           print("Find the sources defining it and add them, " +
                 "or provide a stub.")
       tip = {"message": "Found function with missing spec: " + fname + "\n" +
              "   Looking for files defining it...",
              "action":partial(action, fname)
       }
       tips.append(tip)
    else:
       match = re_recursive_call_start.search(line)
       if match:
          def action():
             print("Consider patching or stubbing the recursive call, " +
                   "then re-run the analysis.")
          msg_lines = []
          line = next(lines)
          while True:
             match = re_recursive_call_end.search(line)
             if match:
                tip = {"message": "Found recursive call at:\n" +
                       "\n".join(msg_lines),
                       "action":action
                       }
                tips.append(tip)
                break
             else:
                msg_lines.append(line)
                try:
                   line = next(lines)
                except StopIteration:
                   print("** Error: EOF without ending recursive call stack?")
                   assert False

if tips != []:
   print("")
   print("***** make-wrapper recommendations *****")
   print("")
   counter = 1
   print("*** recommendation #" + str(counter) + " ***")
   print("")
   for tip in tips:
      if counter > 1:
         print("")
         print("*** recommendation #" + str(counter) + " ***")
      print(str(counter) + ". " + tip["message"])
      counter += 1
      tip["action"]()



##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

import re
import sys
import string
import curses


sensitivity = 0.02

class ResultsFormatter(string.Formatter):
    @staticmethod
    def format_memory(kilobytes):
        if kilobytes < 4096:
            return str(kilobytes) + " kiB"
        megabytes = round(kilobytes / 1024)
        if megabytes < 4096:
            return str(megabytes) + " MiB"
        gigabytes = round(megabytes / 1024)
        return str(gygabytes) + " GiB"

    @staticmethod
    def format_time(seconds):
        if seconds < 10:
            return str(round(seconds,2)) + "s"
        if seconds < 100:
            return str(round(seconds,1)) + "s"
        if seconds < 600:
            return str(round(seconds)) + "s"
        minutes = round(seconds / 60)
        if minutes < 600:
            return str(minutes) + "m"
        hours = round(minutes / 60)
        return str(hours) + "h"

    @staticmethod
    def attribute(value, inverted):
        if value > sensitivity:
            return "@-" if inverted else "@+"
        elif value < -sensitivity:
            return "@+" if inverted else "@-"
        else:
            return "@="

    def get_field(self, field_name, args, kwargs):
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError):
            return None,field_name

    def format_field(self, value, format_spec):
        if value == None:
            return ""
        elif format_spec.startswith('+cmp:'):
            remainder = format_spec.split("+cmp:",1)[1]
            return (self.attribute(value, False) +
                self.format_field(value, remainder) + "@=")
        elif format_spec.startswith('-cmp:'):
            remainder = format_spec.split("-cmp:",1)[1]
            return (self.attribute(value, True) +
                self.format_field(value, remainder) + "@=")
        elif format_spec == 'time':
            return self.format_time(value)
        elif format_spec == 'memory':
            return self.format_memory(value)
        else:
            return super().format_field(value, format_spec)

class UserExitRequest (Exception):
    pass

class PlainDisplay:
    NEGATIVE = 1
    POSITIVE = 2
    RUNNING = 3
    HEADER = 4

    columns = [
        {"size":64, "caption":"Case", "format":"{target_name:s}"},
        {"size":14, "caption":"Coverage",
         "format":"{coverage:>8.0%} {diff_coverage:+cmp:+.0%}"},
        {"size":14, "caption":"Alarms",
         "format":"{alarms:>8d} {diff_alarms:-cmp:+d}"},
        {"size":14 , "caption":"Warnings",
         "format":"{warnings:>8d} {diff_warnings:-cmp:+d}"},
        {"size":14, "caption":"Time",
         "format":"{user_time:time} {diff_user_time:-cmp:+.0%}"},
        {"size":14, "caption":"Memory",
         "format":"{memory:memory} {diff_memory:-cmp:+.0%}"}]

    def __init__(self):
        self.NEGATIVE = 0
        self.POSITIVE = 0
        self.RUNNING = 0
        self.HEADER = 0
        self.OBSOLETE = 0
        self.needs_update = False

    def write(self, text, attributes=0):
        sys.stdout.write(text)

    def rich_write(self, text, override=None, size=0):
        attributes = 0
        n = 0
        for s in re.split(r'(@.)', text):
            if s == "@=":
                attributes = 0
            elif s == "@+":
                attributes = self.POSITIVE
            elif s == "@-":
                attributes = self.NEGATIVE
            else:
                n += len(s)
                self.write(s, attributes if override is None else override)
        if n < size:
            self.write(' ' * (size - n), attributes if override is None else override)

    fmt = ResultsFormatter()

    def format(self, *args, **kwargs):
        return self.fmt.format(*args, **kwargs)

    def print_table(self, results):
        self.write(" ", self.HEADER)
        for column in self.columns:
            self.write(self.format('{caption:^{size}}', **column), self.HEADER)
            self.write(" ", self.HEADER)

        self.write("\n-")
        for column in self.columns:
            self.write(self.format('{:-^{size}}', "", **column))
            self.write("-")

        self.write("\n")
        for result in results:
            self.write(" ")
            for column in self.columns:
                s = self.fmt.format(column["format"], **result)
                if result["is_running"]:
                    attribute = self.RUNNING
                elif not result["up_to_date"]:
                    attribute = self.OBSOLETE
                else:
                    attribute = None
                self.rich_write(s, attribute, size=column['size'])

                self.write(" ")
            self.write("\n")

        self.write("\n")
        self.needs_update = False

    def process_inputs(self):
        pass


class CursesDisplay(PlainDisplay):
    def __init__(self, stdscr):
        self.stdscr = stdscr
        #curses.mousemask(curses.ALL_MOUSE_EVENTS)
        stdscr.nodelay(True)
        stdscr.refresh() # Needs to be done once or nothing will be output

        self.window = curses.newpad(400, 160)
        curses.init_color(curses.COLOR_YELLOW, 300, 300, 300)
        curses.init_pair(1, curses.COLOR_RED, 0)
        curses.init_pair(2, curses.COLOR_GREEN, 0)
        curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_YELLOW)
        curses.init_pair(4, curses.COLOR_YELLOW, 0)
        self.NEGATIVE = curses.color_pair(1)
        self.POSITIVE = curses.color_pair(2)
        self.RUNNING = curses.A_BLINK | curses.color_pair(3)
        self.HEADER = curses.A_BOLD
        self.OBSOLETE = curses.color_pair(4)
        self.scroll_y = 0

    def write(self, text, attributes=0):
        self.window.addstr(text, attributes)

    def print_table(self, results):
        self.window.clear()
        PlainDisplay.print_table(self, results)
        height, width = self.stdscr.getmaxyx()
        try:
            self.window.refresh(0, 0, 0, 0, 1, width-1)
            self.window.refresh(self.scroll_y+2, 0, 2, 0, height-1, width-1)
        except Exception:
            # getmaxyx may be out of date, especially when resizing down the
            # window ; just ignore errors
            pass


    def process_inputs(self):
        previous_y = self.scroll_y
        c = self.stdscr.getch()
        while c != -1:
            if c == ord('q'):
                raise UserExitRequest
            elif c == curses.KEY_UP:
                self.scroll_y -= 1
            elif c == curses.KEY_DOWN:
                self.scroll_y += 1
            elif c == curses.KEY_NPAGE:
                self.scroll_y += 10
            elif c == curses.KEY_PPAGE:
                self.scroll_y -= 10
            elif c == curses.KEY_MOUSE:
                id,x,y,z,bstate = curses.getmouse()
                if z > 0:
                    self.scroll_y += 1
                elif z < 0:
                    self.scroll_y -= 1
            self.scroll_y = max(0, self.scroll_y)
            c = self.stdscr.getch()
        if self.scroll_y != previous_y:
            self.needs_update = True


def wrapper(f, *args, **kwargs):
    if 'curses' in kwargs:
        use_curses = kwargs['curses']
        del kwargs['curses']
    else:
        use_curses = False

    if use_curses:
        def g(stdscr):
            nonlocal f, args, kwargs
            display = CursesDisplay(stdscr)
            return f(display, *args, **kwargs)
        return curses.wrapper(g)
    else:
        display = PlainDisplay()
        return f(display, *args, **kwargs)




#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# This script parses a compile_commands.json[1] file and lists the C files
# in it.
#
# [1] See: http://clang.llvm.org/docs/JSONCompilationDatabase.html

import sys
import os
import json
import re
from pathlib import Path

MIN_PYTHON = (3, 6) # for glob(recursive) and automatic Path conversions
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

if len(sys.argv) < 2:
   # no argument, assume default name
   arg = Path("compile_commands.json")
else:
   arg = Path(sys.argv[1])

if not arg.exists():
   print(f"error: file '{arg}' not found")
   sys.exit(f"usage: {sys.argv[0]} [compile_commands.json]")

# check if arg has a known extension
def is_known_c_extension(ext):
   return ext == ".c" or ext == ".i" or ext == ".h"

pwd = os.getcwd()
fcmake_pwd = pwd / Path(".frama-c") # pwd as seen by the Frama-C makefile
json = json.loads(open(arg).read())
jcdb_dir = arg.parent
includes = set()
defines = set()
files = set() # set of pairs of (file, file_for_fcmake)
for entry in json:
   arg_includes = [] # before normalization
   if not "file" in entry:
       # ignore entries without a filename
       continue
   file = Path(entry["file"])
   dir = Path(entry["directory"]) if "directory" in entry else None
   if file.is_absolute():
      filepath = file
   elif dir and dir.is_absolute():
      filepath = dir / file
   elif dir:
      filepath = jcdb_dir / dir / file
   else:
      filepath = jcdb_dir / file
   if not is_known_c_extension(filepath.suffix):
      print(f"warning: ignoring file of unknown type: {filepath}")
   else:
      files.add((os.path.relpath(filepath, pwd), os.path.relpath(filepath, fcmake_pwd)))

files_for_fcmake = map (lambda x:(x[1]), files)
print("# Paths as seen by a makefile inside subdirectory '.frama-c':")
print("SRCS=\\\n" + " \\\n".join(sorted(files_for_fcmake)) + " \\")
print("")

files_defining_main = set()
re_main = re.compile("(int|void)\s+main\s*\([^)]*\)\s*\{")
for (file, file_for_fcmake) in files:
   assert os.path.exists(file), "file does not exist: %s" % file
   with open(file, 'r') as content_file:
      content = content_file.read()
      res = re.search(re_main, content)
      if res is not None:
         files_defining_main.add(file_for_fcmake)

if files_defining_main != []:
   print("")
   print("# Possible definition of main function in the following file(s), as seen from '.frama-c':")
   print("\n".join(sorted(files_defining_main)))



#!/usr/bin/env python3
#-*- coding: utf-8 -*-
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2020                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# This script is used to interactively fill template.mk, converting it
# into a GNUmakefile ready for analysis.

import sys
import os
import re
import shutil
import shlex
import glob
import json
from subprocess import Popen, PIPE
from pathlib import Path
import function_finder

MIN_PYTHON = (3, 6) # for glob(recursive) and automatic Path conversions
if sys.version_info < MIN_PYTHON:
    sys.exit("Python %s.%s or later is required.\n" % MIN_PYTHON)

if len(sys.argv) > 2:
    print(f"usage: {sys.argv[0]} [dir]")
    print("       creates a Frama-C makefile in [dir] (default: .frama-c)")
    sys.exit(1)

framac_bin = os.getenv('FRAMAC_BIN')
if not framac_bin:
    sys.exit("error: FRAMAC_BIN not in environment (set by frama-c-script)")
framac_bin = Path(framac_bin)

jcdb = Path("compile_commands.json")

dir = Path(sys.argv[1] if len(sys.argv) == 2 else ".frama-c")
fc_stubs_c = dir / "fc_stubs.c"
gnumakefile = dir / "GNUmakefile"

print(f"Preparing template: {gnumakefile}")

# relative prefix, due to GNUmakefile possibly being in a sub-directory of PWD
relprefix = os.path.relpath(os.getcwd(), dir)

if "PTESTS_TESTING" in os.environ:
    print("Running ptests: setting up mock files...")
    jcdb.touch()
    Path(dir).mkdir(parents=True, exist_ok=True)
    fc_stubs_c.touch()
    gnumakefile.touch()

process = Popen([framac_bin / "frama-c", "-print-config-json"], stdout=PIPE)
(output, err) = process.communicate()
exit_code = process.wait()
if exit_code != 0:
    sys.exit(f"error running frama-c -print-config-json")

fc_config = json.loads(output.decode('utf-8'))
sharedir = Path(fc_config['datadir'])
default_machdep = fc_config['current_machdep']
machdeps = fc_config['machdeps']

def check_path_exists(path):
    if path.exists():
        yn = input(f"warning: {path} already exists. Overwrite? [y/N] ")
        if yn == "" or not (yn[0] == "Y" or yn[0] == "y"):
            print("Exiting without overwriting.")
            sys.exit(0)
    pathdir = path.parent
    if not pathdir.exists():
        yn = input(f"warning: directory '{pathdir}' does not exit. Create it? [y/N] ")
        if yn == "" or not (yn[0] == "Y" or yn[0] == "y"):
            print("Exiting without creating.")
            sys.exit(0)
        pathdir.mkdir(parents=True, exist_ok=False)

check_path_exists(gnumakefile)
main = input("Main target name: ")
if not re.match("^[a-zA-Z_0-9-]+$", main):
    sys.exit("error: invalid main target name (can only contain letters, digits, dash or underscore)")

main_fun_finder_re = function_finder.prepare("main")

# returns 0 if none, 1 if declaration, 2 if definition
def defines_or_declares_main(f):
    return function_finder.find(main_fun_finder_re, f)

def expand_and_normalize_sources(expression, relprefix):
    subexps = shlex.split(expression)
    sources_lists = [glob.glob(exp, recursive=True) for exp in subexps]
    sources = sorted(set([item for sublist in sources_lists for item in sublist]))
    return sources

def rel_prefix(f):
    return f"{f}" if os.path.isabs(f) else f"{relprefix}/{f}"

def sources_list_for_makefile(sources):
    return "\n".join([f"  {rel_prefix(source)} \\" for source in sources])

def main_suffix(f):
    main = defines_or_declares_main(f)
    if main == 2:
        return "\t# defines 'main'"
    elif main == 1:
        return "\t# declares 'main'"
    else:
        return ""

while True:
    sources = input("Source files (default: **/*.c): ")
    if not sources:
        sources="**/*.c"
    source_list = expand_and_normalize_sources(sources, relprefix)
    if not source_list:
        print(f"error: no sources were matched for '{sources}'.")
    else:
        print(f"The following sources were matched (relative to {dir}):")
        print("\n".join(["  " + rel_prefix(source) + main_suffix(source) for source in source_list]))
        print()
        definitions_of_main = len([source for source in source_list if defines_or_declares_main(source)])
        if definitions_of_main > 1:
            print("warning: 'main' seems to be defined multiple times.")
        yn = input("Is this ok? [Y/n] ")
        if yn == "" or not (yn[0] == "N" or yn[0] == "n"):
            break

json_compilation_database = None
if jcdb.is_file():
    yn = input("compile_commands.json exists, add option -json-compilation-database? [Y/n] ")
    if yn == "" or not (yn[0] == "N" or yn[0] == "n"):
        json_compilation_database = f"{relprefix}/compile_commands.json"
    else:
        print("Option not added; you can later add it to FCFLAGS.")

add_main_stub = False
yn = input("Add stub for function main (only needed if it uses command-line arguments)? [y/N] ")
if yn != "" and (yn[0] == "Y" or yn[0] == "y"):
    add_main_stub = True

print("Please define the architectural model (machdep) of the target machine.")
print("Known machdeps: " + " ".join(machdeps))
machdep_chosen = False
while not machdep_chosen:
    machdep = input(f"Please enter the machdep [{default_machdep}]: ")
    if not machdep:
        machdep = default_machdep
        machdep_chosen = True
    else:
        if not (machdep in machdeps):
            yn = input(f"'{machdep}' is not a standard machdep. Proceed anyway? [y/N]")
            if yn != "" and (yn[0] == "Y" or yn[0] == "y"):
                machdep_chosen = True
        else:
            machdep_chosen = True

def insert_line_after(lines, line_pattern, newline):
    re_line = re.compile(line_pattern)
    for i in range(0, len(lines)):
        if re_line.search(lines[i]):
            lines.insert(i+1, newline)
            return lines
    sys.exit(f"error: no lines found matching pattern: {line_pattern}")

def replace_line(lines, line_pattern, value, all_occurrences=False):
    replaced = False
    re_line = re.compile(line_pattern)
    for i in range(0, len(lines)):
        if re_line.search(lines[i]):
            lines[i] = value
            replaced = True
            if not all_occurrences:
                return lines
    if replaced:
        return lines
    else:
        sys.exit(f"error: no lines found matching pattern: {line_pattern}")

def remove_lines_between(lines, start_pattern, end_pattern):
    re_start = re.compile(start_pattern)
    re_end = re.compile(end_pattern)
    first_to_remove = -1
    last_to_remove = -1
    for i in range(0, len(lines)):
        if first_to_remove == -1 and re_start.search(lines[i]):
            first_to_remove = i
        elif re_end.search(lines[i]):
            last_to_remove = i
            break
    if first_to_remove == -1:
        sys.exit("error: could not find start pattern: " + start_pattern)
    elif last_to_remove == -1:
        sys.exit("error: could not find end pattern: " + end_pattern)
    return (lines[:first_to_remove-1] if first_to_remove > 0 else []) + (lines[last_to_remove+1:] if last_to_remove < len(lines)-1 else [])

with open(sharedir / "analysis-scripts" / "template.mk") as f:
    lines = list(f)
    lines = replace_line(lines, "^MACHDEP = .*", f"MACHDEP = {machdep}\n")
    if add_main_stub:
        lines = insert_line_after(lines, "^main.parse: \\\\", f"  fc_stubs.c \\\n")
        check_path_exists(fc_stubs_c)
        shutil.copyfile(sharedir / "analysis-scripts" / "fc_stubs.c", fc_stubs_c)
        lines = insert_line_after(lines, "^FCFLAGS", "  -main eva_main \\\n")
        print(f"Created stub for main function: {dir / 'fc_stubs.c'}")
    lines = replace_line(lines, "^main.parse: \\\\", f"{main}.parse: \\\n")
    lines = replace_line(lines, "^TARGETS = main.eva", f"TARGETS = {main}.eva\n")
    lines = replace_line(lines, "^  main.c \\\\", sources_list_for_makefile(source_list) + "\n")
    if json_compilation_database:
      lines = insert_line_after(lines, "^FCFLAGS", f"  -json-compilation-database {json_compilation_database} \\\n")
    if relprefix != "..":
        lines = replace_line(lines, "^  -add-symbolic-path=.:.. \\\\", f"  -add-symbolic-path=.:{relprefix} \\\n", all_occurrences=True)

gnumakefile.write_text("".join(lines))

print(f"Template created: {gnumakefile}")

# write path.mk
path_mk = dir / "path.mk"

with open(path_mk, "w") as f:
    f.write(f"""FRAMAC_BIN={framac_bin}
ifeq ($(wildcard $(FRAMAC_BIN)),)
# Frama-C not installed locally; using the version in the PATH
else
FRAMAC=$(FRAMAC_BIN)/frama-c
FRAMAC_GUI=$(FRAMAC_BIN)/frama-c-gui
endif
""")

print(f"Path to Frama-C binaries written to: {path_mk}")

if "PTESTS_TESTING" in os.environ:
    print("Running ptests: cleaning up after tests...")
    jcdb.unlink()
    fc_stubs_c.unlink()
    path_mk.unlink()
    # gnumakefile is not erased because we want it as an oracle
